@@ -3,6 +3,9 @@ import warnings
 from contextlib import contextmanager
 from IPython.display import Image, display
 import pandas as pd
+import numpy as np
+import re
+import plotly.express as px
 
 with warnings.catch_warnings():
     warnings.simplefilter("ignore")
@@ -19,17 +22,19 @@ with warnings.catch_warnings():
 
 NumberTypes = (int, float, complex)
 
+
 @contextmanager
-def __r_inline_plot(width=600, height=600, dpi=100):
+def __r_inline_plot(width=600, height=520, dpi=150, plot_it=True):
     
     """
     Display R plots inline.
 
     Parameters
     ----------
-    width, height : int or float, default 600
+    width, height : numeric, default 600 by 520
         Width and height of the plot.
-    dpi : int or float, default 100
+        
+    dpi : numeric, default 150
         Resolution of the plot.
     """
     
@@ -41,9 +46,205 @@ def __r_inline_plot(width=600, height=600, dpi=100):
         yield
 
     data = b.getvalue()
-    display(Image(data=data, format='png', embed=True))
+    if plot_it:
+        display(Image(data=data, format='png', embed=True))
 
     
+def __flatten_list(_2d_list):
+    flat_list = []
+    # Iterate through the outer list
+    for element in _2d_list:
+        if type(element) is list:
+            # If the element is of type list, iterate through the sublist
+            for item in element:
+                flat_list.append(item)
+        else:
+            flat_list.append(element)
+    return flat_list
+
+
+def __format_chemical_name(name):
+    p = re.compile(r'(?P<sp>\+|-\d+?$)')
+    name = p.sub(r'<sup>\g<sp></sup>', name)
+    
+    name_no_charge = re.match(r'(?:(?!<|$).)*', name).group(0)
+    mapping = {"0": "<sub>0</sub>", "1": "<sub>1</sub>", "2": "<sub>2</sub>", "3": "<sub>3</sub>", "4": "<sub>4</sub>", 
+           "5": "<sub>5</sub>", "6": "<sub>6</sub>", "7": "<sub>7</sub>", "8": "<sub>8</sub>", "9": "<sub>9</sub>",
+           ".":"<sub>.</sub>"}
+    name_no_charge_formatted = "".join([mapping.get(x) or x for x in list(name_no_charge)])
+    name = re.sub(name_no_charge, name_no_charge_formatted, name)
+
+    return(name)
+
+
+def diagram_interactive(data, balance=None, width=600, height=520, alpha=False, messages=True):
+    
+    """
+    Produce an interactive Plotly plot.
+    
+    Parameters
+    ----------
+    data : rpy2.ListVector
+        Output from `equilibrate` or `affinity`.
+
+    alpha : bool or str (balance), default False
+        For speciation diagrams, plot degree of formation instead of
+        activities?
+
+    balance : str or numeric, optional
+        How to balance the transformations.
+    
+    width, height : numeric, default 600 by 520
+        Width and height of the plot.
+    
+    messages : bool, default True
+        Display messages from CHNOSZ?
+    
+    interactive : bool, default False
+        Experimental! Display an interactive plot?
+    
+    Returns
+    -------
+    Produces an interactive Plotly plot.
+    """
+    
+    basis_sp = data.rx2("basis").rownames
+
+    xyvars = list(data.rx2("vars"))
+    xyvals = list(data.rx2("vals"))
+
+    if 'loga.equil' not in data.names:
+        calc_type = "a"
+    else:
+        calc_type = "e"
+
+    data = equilibrate(data, balance=balance, messages=messages)
+
+    if calc_type=="a":
+        out_vals = data.rx2("values")
+        out_units = "A/(2.303RT)"
+        df = pandas2ri.ri2py_dataframe(out_vals)
+        df["n.balance"] = list(data.rx2("n.balance"))
+        
+        # divide values by balance
+        df = df.apply(lambda row: row/row["n.balance"], axis=1)
+        df = df.drop(["n.balance"], axis=1)
+        sp = list(info([int(val) for val in list(out_vals.names)], messages=False)["name"])
+        
+    elif calc_type=="e":
+        out_vals = data.rx2("loga.equil")
+        out_units = "log a"
+        df = pandas2ri.ri2py_dataframe(out_vals)
+        sp = list(info([int(val) for val in list(data.rx2("values").names)], messages=False)["name"])
+    
+    df.index = sp
+    df = df.transpose()
+
+    if alpha and len(xyvars) == 1:
+        df = df.applymap(lambda x: 10**x)
+        df = df[sp].div(df[sp].sum(axis=1), axis=0)
+        
+    xvar = xyvars[0]
+    xvals = [float(val) for val in xyvals[0]]
+
+    if len(xyvars) == 1:
+        df[xvar] = xvals
+        if alpha:
+            ylabel = "alpha"
+        else:
+            ylabel = out_units
+        df = pd.melt(df, id_vars=xyvars, value_vars=sp)
+
+    elif len(xyvars)==2:
+        # predominance plot
+        yvar = xyvars[1]
+        yvals = [float(val) for val in xyvals[1]]
+        df["pred"] = df.idxmax(axis=1)
+        df["prednames"] = df["pred"]
+
+        xvals_full = xvals*len(yvals)
+        yvals_full = __flatten_list([[y]*len(xvals) for y in yvals])
+        df[xvar] = xvals_full
+        df[yvar] = yvals_full
+
+    unit_dict = {"P":"bar", "T":"°C", "pH":"", "Eh":"volts", "IS":"mol/kg"}
+
+    for s in basis_sp:
+        unit_dict[s] = "logact "+s
+
+    xlab = xvar+", "+unit_dict[xvar]
+    if xvar == "pH":
+        xlab = "pH"
+
+    if xvar in basis_sp:
+        xlab = unit_dict[xvar]
+
+    if len(xyvars) == 1:
+        
+        fig = px.line(df, x=xvar, y="value", color='variable', template="simple_white",
+                      width=width,  height=height,
+                      labels=dict(value=ylabel, x=__format_chemical_name(xlab)),
+                     )
+        fig.update_layout(xaxis_title=__format_chemical_name(xlab),
+                          yaxis_title=ylabel,
+                          )
+    
+        config = {'displaylogo': False,
+                  'modeBarButtonsToRemove': ['resetScale2d', 'toggleSpikelines']}
+        fig.show(config=config)
+
+    if len(xyvars) == 2:
+        mappings = {'pred': {s:lab for s,lab in zip(sp,range(0,len(sp)))}}
+        df.replace(mappings, inplace=True)
+
+        data = np.array(df.pred)
+        shape = (len(xvals), len(yvals))
+        dmap = data.reshape(shape)
+
+        data = np.array(df.prednames)
+        shape = (len(xvals), len(yvals))
+        dmap_names = data.reshape(shape)
+
+        ylab = yvar+", "+unit_dict[yvar]
+        if yvar in basis_sp:
+            ylab = unit_dict[yvar]
+
+        if yvar == "pH":
+            yvar = "pH"
+
+        fig = px.imshow(dmap, width=width, height=height, aspect="auto",
+                        labels=dict(x=__format_chemical_name(xlab), y=__format_chemical_name(ylab), color="region"),
+                        x=xvals, y=yvals, template="simple_white",
+                       )
+
+        fig.update(data=[{'customdata': dmap_names,
+            'hovertemplate': xvar+': %{x} '+unit_dict[xvar]+'<br>'+yvar+': %{y} '+unit_dict[yvar]+'<br>Region: %{customdata}<extra></extra>'}])
+
+        fig.update_traces(dict(showscale=False, 
+                               coloraxis=None, 
+                               colorscale='viridis'),
+                          selector={'type':'heatmap'})
+
+        fig.update_yaxes(autorange=True)
+
+        for s in sp:
+            if s in set(df["prednames"]):
+                df_s = df.loc[df["prednames"]==s,]
+                namex = df_s[xvar].mean()
+                namey = df_s[yvar].mean()
+                fig.add_annotation(x=namex, y=namey,
+                                   text=__format_chemical_name(s),
+                                   bgcolor="rgba(255, 255, 255, 0.5)",
+                                   showarrow=False)
+
+        config = {'displaylogo': False,
+                  'modeBarButtonsToRemove': ['zoom2d', 'pan2d', 'zoomIn2d', 'zoomOut2d',
+                                             'autoScale2d', 'resetScale2d', 'toggleSpikelines',
+                                             'hoverClosestCartesian', 'hoverCompareCartesian']}
+
+        fig.show(config=config)
+
+
 def _convert_to_RVector(value, force_Rvec=True):
     
     """
@@ -81,6 +282,73 @@ def _convert_to_RVector(value, force_Rvec=True):
     else:
         return ro.StrVector(value)
             
+
+def water(property=None, T=298.15, P="Psat", P1=True, messages=True):
+    
+    """
+    Python wrapper for the water() function in CHNOSZ.
+    Calculate thermodynamic and electrostatic properties of water.
+    
+    Parameters
+    ----------
+    property : str or list of str, optional
+        Computational setting or property(s) to calculate. To set a water model,
+        use 'SUPCRT92' (default) or 'SUPCRT', 'IAPWS95' or 'IAPWS', or 'DEW'.
+        To calculate a property, use 'A', 'G', 'S', 'U', etc. See 
+        http://chnosz.net/manual/water.html for the complete catalog of
+        properties that can be calculated, their units, and their availability
+        in water models.
+
+    T : numeric, default 298.15
+        Temperature (K)
+
+    P : numeric, default "Psat"
+        Pressure (bar), or Psat for vapor-liquid saturation.
+
+    P1 : bool, default True
+        Output pressure of 1 bar below 100 °C instead of calculated values of Psat?
+
+    messages : bool, default True
+        Display messages from CHNOSZ?
+
+    Returns
+    ----------
+    out : float or dict
+        Calculated value of desired water property. If `property` is a list,
+        returns a dictionary of calculated values.
+    """
+    
+    if property == None:
+        property = ro.r("NULL")
+    elif isinstance(property, list):
+        property = _convert_to_RVector(property, force_Rvec=True)
+    else:
+        pass
+    
+    args = {'property':property, 'T':T, 'P':P, 'P1':P1}
+    
+    with warnings.catch_warnings(record=True) as w:
+        warnings.simplefilter("always")
+        out = CHNOSZ.water(**args)
+
+    if messages:
+        for warning in w:
+            print(warning.message)
+    
+    if property not in ["SUPCRT92", "SUPCRT", "IAPWS95", "IAPWS", "DEW"]:
+        out = list(out)
+        if len(out) == 1:
+            if property == ro.r("NULL"):
+                out = out[0]
+            else:
+                out = out[0][0]
+        else:
+            # CHNOSZ produces a single-row dataframe when multiple properties
+            # are supplied to `property`. Here, a dictionary is returned.
+            out = {prop:o[0] for prop,o in zip(property, out)}
+        return out
+        
+        
 def entropy(formula, messages=True):
     
     """
@@ -160,7 +428,7 @@ def mass(formula, messages=True):
     return out
 
 
-def ZC(formula, messages=True):
+def zc(formula, messages=True):
     
     """
     Python wrapper for the ZC() function in CHNOSZ.
@@ -406,7 +674,8 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
             dx=0, dy=0, srt=0, min_area=0,
             main=None, legend_x=None,
             add=False, plot_it=True, tplot=True,
-            messages=True):
+            width=600, height=520, dpi=150,
+            messages=True, interactive=False):
     
     """
     Python wrapper for the diagram() function in CHNOSZ.
@@ -529,13 +798,13 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
     family : str, optional
         Font family for names.
 
-    adj : numeric, default 0.5
+    adj : numeric or list, default 0.5
         Adjustment for line labels.
 
-    dx : numeric, default 0
+    dx : numeric or list, default 0
         X offset for line or field labels.
 
-    dy : numeric, default 0
+    dy : numeric or list, default 0
         Y offset for line or field labels.
 
     srt : numeric, default 0
@@ -560,9 +829,18 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
     tplot : bool, default True
         Set up plot with thermo.plot.new?
     
+    width, height : numeric, default 600 by 520
+        Width and height of the plot.
+        
+    dpi : numeric, default 150
+        Resolution of the plot.
+    
     messages : bool, default True
         Display messages from CHNOSZ?
     
+    interactive : bool, default False
+        Experimental! Display an interactive plot?
+    
     Returns
     -------
     a : rpy2.ListVector
@@ -571,6 +849,13 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
         Dictionary of arguments supplied to `diagram`.
     """
     
+    if interactive:
+        diagram_interactive(data=eout, balance=balance,
+                            width=width, height=height,
+                            alpha=alpha, messages=messages)
+        return
+    
+    
     args = {'eout':eout, 'ptype':ptype, 'alpha':alpha, 'normalize':normalize,
             'as.residue':as_residue, 'ylog':ylog, 'fill.NA':fill_NA, 'format.names':format_names,
             'bold':bold, 'italic':italic, 'adj':adj, 'dx':dx, 'dy':dy, 'srt':srt,
@@ -602,12 +887,18 @@ def diagram(eout, ptype='auto', alpha=False, normalize=False,
     if names != None: args["names"] = _convert_to_RVector(names)
     if font != None: args["font"] = _convert_to_RVector(font)
     if family != None: args["family"] = _convert_to_RVector(family)
+    if isinstance(adj, list):
+        args["adj"] = _convert_to_RVector(adj)
+    if isinstance(dx, list):
+        args["dx"] = _convert_to_RVector(dx)
+    if isinstance(dy, list):
+        args["dy"] = _convert_to_RVector(dy)
     if main != None: args["main"] = main
     if legend_x != None: args["legend.x"] = legend_x
     
     with warnings.catch_warnings(record=True) as w:
         warnings.simplefilter("always")
-        with __r_inline_plot(width=1024, height=896, dpi=150):
+        with __r_inline_plot(width=width, height=height, dpi=dpi, plot_it=plot_it):
             if isinstance(add, bool):
                 if add: # add='True' does not work with the current pyCHNOSZ framework
                     raise Exception("The argument 'add' must be assigned the output of the previous diagram(s).")
@@ -1274,4 +1565,4 @@ class thermo(object):
             setattr(self, name, attr)
             
     def __getitem__(self, item):
-         return getattr(self, item)
\ No newline at end of file
+         return getattr(self, item)
